// Copyright (c) ONNX Project Contributors

/*
 * SPDX-License-Identifier: Apache-2.0
 */

#pragma once

namespace ONNX_NAMESPACE {

// Represents the most recent release version. Updated with every release.
constexpr const char* LAST_RELEASE_VERSION = "1.16.0";

} // namespace ONNX_NAMESPACE
